"use strict";

exports.__esModule = true;
exports.default = void 0;
var _index = _interopRequireDefault(require("../parse/index.cjs"));
var _arrayDash = _interopRequireDefault(require("../test/array-dash.cjs"));
var _arrayIndex = _interopRequireDefault(require("../test/array-index.cjs"));
var _TraceBuilder = _interopRequireDefault(require("./trace/TraceBuilder.cjs"));
var _index2 = _interopRequireDefault(require("./realms/json/index.cjs"));
var _JSONPointerEvaluateError = _interopRequireDefault(require("../errors/JSONPointerEvaluateError.cjs"));
var _JSONPointerTypeError = _interopRequireDefault(require("../errors/JSONPointerTypeError.cjs"));
var _JSONPointerIndexError = _interopRequireDefault(require("../errors/JSONPointerIndexError.cjs"));
var _JSONPointerKeyError = _interopRequireDefault(require("../errors/JSONPointerKeyError.cjs"));
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
const evaluate = (value, jsonPointer, {
  strictArrays = true,
  strictObjects = true,
  realm = new _index2.default(),
  trace = false
} = {}) => {
  const {
    result: parseResult,
    tree: referenceTokens,
    trace: parseTrace
  } = (0, _index.default)(jsonPointer, {
    trace: !!trace
  });
  const tracer = trace ? new _TraceBuilder.default(trace, {
    jsonPointer,
    referenceTokens,
    strictArrays,
    strictObjects,
    realm,
    value
  }) : null;
  if (!parseResult.success) {
    let message = `Invalid JSON Pointer: "${jsonPointer}". Syntax error at position ${parseResult.maxMatched}`;
    message += parseTrace ? `, expected ${parseTrace.inferExpectations()}` : '';
    tracer == null || tracer.step({
      referenceToken: undefined,
      input: value,
      success: false,
      reason: message
    });
    throw new _JSONPointerEvaluateError.default(message, {
      jsonPointer
    });
  }
  try {
    let output;
    return referenceTokens.reduce((current, referenceToken, referenceTokenPosition) => {
      if (realm.isArray(current)) {
        if ((0, _arrayDash.default)(referenceToken)) {
          if (strictArrays) {
            const message = `Invalid array index "-" at position ${referenceTokenPosition} in "${jsonPointer}". The "-" token always refers to a nonexistent element during evaluation`;
            tracer == null || tracer.step({
              referenceToken,
              input: current,
              success: false,
              reason: message
            });
            throw new _JSONPointerIndexError.default(message, {
              jsonPointer,
              referenceTokens,
              referenceToken,
              referenceTokenPosition,
              currentValue: current,
              realm: realm.name
            });
          } else {
            output = realm.evaluate(current, String(realm.sizeOf(current)));
            tracer == null || tracer.step({
              referenceToken,
              input: current,
              output
            });
            return output;
          }
        }
        if (!(0, _arrayIndex.default)(referenceToken) && strictArrays) {
          const message = `Invalid array index "${referenceToken}" at position ${referenceTokenPosition} in "${jsonPointer}": index MUST be "0", or digits without a leading "0"`;
          tracer == null || tracer.step({
            referenceToken,
            input: current,
            success: false,
            reason: message
          });
          throw new _JSONPointerIndexError.default(message, {
            jsonPointer,
            referenceTokens,
            referenceToken,
            referenceTokenPosition,
            currentValue: current,
            realm: realm.name
          });
        }
        const index = Number(referenceToken);
        if (index >= realm.sizeOf(current) && strictArrays) {
          const message = `Invalid array index "${index}" at position ${referenceTokenPosition} in "${jsonPointer}": out of bounds`;
          tracer == null || tracer.step({
            referenceToken,
            input: current,
            success: false,
            reason: message
          });
          throw new _JSONPointerIndexError.default(message, {
            jsonPointer,
            referenceTokens,
            referenceToken: index,
            referenceTokenPosition,
            currentValue: current,
            realm: realm.name
          });
        }
        output = realm.evaluate(current, referenceToken);
        tracer == null || tracer.step({
          referenceToken,
          input: current,
          output
        });
        return output;
      }
      if (realm.isObject(current)) {
        if (!realm.has(current, referenceToken) && strictObjects) {
          const message = `Invalid object key "${referenceToken}" at position ${referenceTokenPosition} in "${jsonPointer}": key not found in object`;
          tracer == null || tracer.step({
            referenceToken,
            input: current,
            success: false,
            reason: message
          });
          throw new _JSONPointerKeyError.default(message, {
            jsonPointer,
            referenceTokens,
            referenceToken,
            referenceTokenPosition,
            currentValue: current,
            realm: realm.name
          });
        }
        output = realm.evaluate(current, referenceToken);
        tracer == null || tracer.step({
          referenceToken,
          input: current,
          output
        });
        return output;
      }
      const message = `Invalid reference token "${referenceToken}" at position ${referenceTokenPosition} in "${jsonPointer}": cannot be applied to a non-object/non-array value`;
      tracer == null || tracer.step({
        referenceToken,
        input: current,
        success: false,
        reason: message
      });
      throw new _JSONPointerTypeError.default(message, {
        jsonPointer,
        referenceTokens,
        referenceToken,
        referenceTokenPosition,
        currentValue: current,
        realm: realm.name
      });
    }, value);
  } catch (error) {
    if (error instanceof _JSONPointerEvaluateError.default) {
      throw error;
    }
    throw new _JSONPointerEvaluateError.default('Unexpected error during JSON Pointer evaluation', {
      cause: error,
      jsonPointer,
      referenceTokens
    });
  }
};
var _default = exports.default = evaluate;